package de.hftstuttgart.dtabackend.utils;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.net.MalformedURLException;
import java.nio.file.Path;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import com.fasterxml.jackson.core.exc.StreamReadException;
import com.fasterxml.jackson.databind.DatabindException;

import de.hftstuttgart.dtabackend.models.ExerciseCompetencyProfile;
import de.hftstuttgart.dtabackend.models.ICompetencyProfile;
import de.hftstuttgart.dtabackend.models.Result;
import de.hftstuttgart.dtabackend.models.ResultSummary;
import de.hftstuttgart.dtabackend.models.TestCompetencyProfile;

import java.io.FileNotFoundException;

public class CompetencyAssessmentUtil {
	private static final Logger LOG = LogManager.getLogger(CompetencyAssessmentUtil.class);

	public static String TEST_COMPETENCY_MANIFEST_FILE_NAME = "competency-tests.mft";
	public static String EXERCISE_COMPETENCY_MANIFEST_FILE_NAME = "exercise-tests.mft";

	public static float[] sumTestCompetencyProfiles(List<TestCompetencyProfile> testCompetencyProfiles) {
		LOG.debug("Starting sumTestCompetencyProfiles with {} test profiles.", testCompetencyProfiles.size());

		float[] tcpTotalProfile = new float[ICompetencyProfile.MAX_COMPETENCY_DIMENSIONS];
		for (TestCompetencyProfile currentProfile : testCompetencyProfiles) {
			LOG.debug("Adding competency assessments from profile: {}", currentProfile.testName);
			tcpTotalProfile = ICompetencyProfile.competencySum(tcpTotalProfile, currentProfile.competencyAssessments);
		}

		LOG.debug("Completed summing test competency profiles. Total profile: {}", packFloats(tcpTotalProfile));
		return tcpTotalProfile;
	}

	public static float[] sumSuccessfulCompetencyProfiles(List<TestCompetencyProfile> testCompetencyProfiles, ResultSummary resultSummary, boolean success) {
		LOG.debug("Starting sumSuccessfulCompetencyProfiles. Success flag: {}", success);

		float[] sumSuccessful = new float[ICompetencyProfile.MAX_COMPETENCY_DIMENSIONS];
		for (Result currentResult : resultSummary.results) {
			boolean isSuccess = Integer.valueOf(currentResult.state).equals(Result.State.SUCCESS.ordinal());
			LOG.debug("Processing result: {} with success state: {}", currentResult.name, isSuccess);

			if (isSuccess == success) {
				TestCompetencyProfile currentProfile = new TestCompetencyProfile();
				currentProfile.testPackageName = (currentResult.packageName != null) ? currentResult.packageName : "";
				currentProfile.testClassName = (currentResult.className != null) ? currentResult.className : "";
				currentProfile.testName = (currentResult.name != null) ? currentResult.name : "";

				int testIndex = testCompetencyProfiles.indexOf(currentProfile);
				if (testIndex != -1) {
					LOG.debug("Found matching profile for result: {}. Adding competencies.", currentResult.name);
					sumSuccessful = ICompetencyProfile.competencySum(sumSuccessful, testCompetencyProfiles.get(testIndex).competencyAssessments);
				} else {
					LOG.debug("No matching profile found for result: {}", currentResult.name);
				}
			}
		}

		LOG.debug("Completed summing successful competency profiles. Sum: {}", packFloats(sumSuccessful));
		return sumSuccessful;
	}

	public static List<TestCompetencyProfile> readTestCompetencyProfiles(Path testPath, String fileName) {
		LOG.debug("Reading test competency profiles from path: {}, file: {}", testPath, fileName);

		List<TestCompetencyProfile> testCompetencyProfiles = new ArrayList<>();
		try (BufferedReader testCompetencyManifest = new BufferedReader(new FileReader(new File(testPath.toFile(), fileName)))) {
			String testEntry = testCompetencyManifest.readLine();

			while (testEntry != null) {
				String[] testEntyComponents = testEntry.split(ICompetencyProfile.COMPETENCY_SEPARATOR);
				TestCompetencyProfile currentProfile = new TestCompetencyProfile();

				currentProfile.testPackageName = testEntyComponents[0];
				currentProfile.testClassName = testEntyComponents[1];
				currentProfile.testName = testEntyComponents[2];
				for (int competencyIndex = 0; competencyIndex < ICompetencyProfile.MAX_COMPETENCY_DIMENSIONS; competencyIndex++) {
					currentProfile.competencyAssessments[competencyIndex] = Float.valueOf(testEntyComponents[competencyIndex + 3]);
				}

				testCompetencyProfiles.add(currentProfile);
				LOG.debug("Added test competency profile: {}", currentProfile.testName);

				testEntry = testCompetencyManifest.readLine();
			}

			LOG.info("Added {} test competency profiles from manifest.", testCompetencyProfiles.size());
		} catch (FileNotFoundException e) {
			LOG.info("Test competency manifest file not found. Skipping functionality.");
		} catch (IOException e) {
			LOG.info("Test competency manifest file unreadable. Skipping functionality.");
		}

		return testCompetencyProfiles;
	}

	public static List<ExerciseCompetencyProfile> readExerciseCompetencyProfiles(Path exercisePath, String fileName) {
		LOG.debug("Reading exercise competency profiles from path: {}, file: {}", exercisePath, fileName);

		List<ExerciseCompetencyProfile> exerciseCompetencyProfiles = new ArrayList<>();

		try (BufferedReader exerciseCompetencyManifest = new BufferedReader(new FileReader(new File(exercisePath.toFile(), fileName)))) {
			String exerciseEntry = exerciseCompetencyManifest.readLine();

			while (exerciseEntry != null) {
				String[] exerciseEntyComponents = exerciseEntry.split(ExerciseCompetencyProfile.COMPETENCY_SEPARATOR);
				ExerciseCompetencyProfile currentProfile = new ExerciseCompetencyProfile();

				currentProfile.exerciseTopicName = exerciseEntyComponents[0];
				currentProfile.exerciseName = exerciseEntyComponents[1];
				currentProfile.exerciseURL = exerciseEntyComponents[2];

				for (int competencyIndex = 0; competencyIndex < ExerciseCompetencyProfile.MAX_COMPETENCY_DIMENSIONS; competencyIndex++) {
					currentProfile.competencyAssessments[competencyIndex] = Float.valueOf(exerciseEntyComponents[competencyIndex + 3]);
				}

				currentProfile.difficulty = Float.parseFloat(exerciseEntyComponents[19]);

				exerciseCompetencyProfiles.add(currentProfile);
				LOG.debug("Added exercise competency profile: {}", currentProfile.exerciseName);

				exerciseEntry = exerciseCompetencyManifest.readLine();
			}

			LOG.info("Added {} exercise competency profiles from manifest.", exerciseCompetencyProfiles.size());
		} catch (FileNotFoundException e) {
			LOG.info("Exercise competency manifest file not found.");
		} catch (IOException e) {
			LOG.info("Exercise competency manifest file unreadable.");
		}

		return exerciseCompetencyProfiles;
	}

	public static String packFloats(float[] array) {
		LOG.debug("Packing float array into string: {}", array);
		return IntStream.range(0, array.length)
				.mapToObj(i -> String.valueOf(array[i]))
				.collect(Collectors.joining(";"));
	}
}
